# /*
#  *  Copyright 2018 Kjell Winblad (kjellwinblad@gmail.com, http://winsh.me)
#  *
#  *  This file is part of LFCA tree tests.
#  *
#  *  catrees is free software: you can redistribute it and/or modify
#  *  it under the terms of the GNU General Public License as published by
#  *  the Free Software Foundation, either version 3 of the License, or
#  *  (at your option) any later version.
#  *
#  *  catrees is distributed in the hope that it will be useful,
#  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  *  GNU General Public License for more details.
#  *
#  *  You should have received a copy of the GNU General Public License
#  *  along with catrees.  If not, see <http://www.gnu.org/licenses/>.
#  */


from subprocess import call
import re

configuration = {
    'input_file': 'lock_free_catree.c',
    'difftool': 'meld',
    'review_transformations': False,

    'pick_out_pattern': r"/\*PICK_OUT_START:([a-zA-Z0-9_]*)\*/\s*\n(.*?)(?=\n\s*/\*PICK_OUT_END\*/)",
    'replace_pattern': r"(\n|^)([^\n]*)/\*REPLACE_START\*/(.*?)(?=/\*REPLACE_END\*/)/\*REPLACE_END\*/\s*\n/\*REPLACEMENT_START\s*\n(.*?)\n(?=(?:/\*|)REPLACEMENT_END\*/)(?:/\*|)REPLACEMENT_END\*/",
    'remove_line_pattern': r"(?=(?:\n|^)(([^\n]*?(?=/\*REMOVE_LINE\*/))/\*REMOVE_LINE\*/\n))",
    'replace_function_calls': [
        {'function_name': 'treap_range_query',
         'new_function_name': 'treap_query',
         'new_number_of_parameters': 5},
        {'function_name': 'all_in_range',
         'new_function_name': 'all_in_range',
         'new_number_of_parameters': 4},
        {'function_name': 'lfcatree_range_query',
         'new_function_name': 'query',
         'new_number_of_parameters': 5},
        {'function_name': 'CAS_rs_linearization',
         'new_function_name': 'CAS',
         'new_number_of_parameters': 3},
        {'function_name': 'CAS_uintptr_t',
         'new_function_name': 'CAS',
         'new_number_of_parameters': 3},
        {'function_name': 'treap_is_empty',
         'new_function_name': 'empty',
         'new_number_of_parameters': 1},
        {'function_name': 'treap_max_key',
         'new_function_name': 'max',
         'new_number_of_parameters': 1},
        {'function_name': 'CAS_store_linearization_points',
         'new_function_name': 'CAS',
         'new_number_of_parameters': 3},
        {'function_name': 'CAS_store_linearization_point',
         'new_function_name': 'CAS',
         'new_number_of_parameters': 3},
        {'function_name': 'leftmost_base_node',
         'new_function_name': 'leftmost',
         'new_number_of_parameters': 1},
        {'function_name': 'try_secure_join_left',
         'new_function_name': 'secure_join_left',
         'new_number_of_parameters': 3},
        {'function_name': 'my_atomic_load',
         'new_function_name': 'aload',
         'new_number_of_parameters': 1},
        {'function_name': 'my_atomic_store',
         'new_function_name': 'astore',
         'new_number_of_parameters': 2},
        {'function_name': 'atomic_store',
         'new_function_name': 'astore',
         'new_number_of_parameters': 2},
        {'function_name': 'try_replace_check_with_model',
         'new_function_name': 'try_replace',
         'new_number_of_parameters': 3},
        {'function_name': 'find_base_node_linearization_point',
         'new_function_name': 'find_base_node',
         'new_number_of_parameters': 2},
        {'function_name': 'treap_lookup_check',
         'new_function_name': 'treap_lookup',
         'new_number_of_parameters': 2},
        {'function_name': 'lfcatree_insert',
         'new_function_name': 'insert',
         'new_number_of_parameters': 2},
        {'function_name': 'lfcatree_remove',
         'new_function_name': 'remove',
         'new_number_of_parameters': 2},
        {'function_name': 'lfcatree_lookup',
         'new_function_name': 'lookup',
         'new_number_of_parameters': 2}


        
    ],
    'to_infix_fields': [
        {'function_name': 'valid_ptrptr',
         'field_name': 'valid',
         'prefix': '&'},
        {'function_name': 'valid',
         'field_name': 'valid)',
         'prefix': 'aload(&'},
        {'function_name': 'status_ptrptr',
         'field_name': 'status',
         'prefix': '&'},
        {'function_name': 'left_ptrptr',
         'field_name': 'left',
         'prefix': '&'},
        {'function_name': 'right_ptrptr',
         'field_name': 'right',
         'prefix': '&'},
        {'function_name': 'join_id_ptrptr',
         'field_name': 'join_id',
         'prefix': '&'},
        {'function_name': 'status_ptrptr',
         'field_name': 'status',
         'prefix': '&'},
        {'function_name': 'joined_neighbor_ptrptr',
         'field_name': 'neigh2',
         'prefix': '&'},
        {'function_name': 'parent',
         'field_name': 'parent',
         'prefix': ''},
        {'function_name': 'grand_parent',
         'field_name': 'gparent',
         'prefix': ''},
        {'function_name': 'other_branch',
         'field_name': 'otherb',
         'prefix': ''},
        {'function_name': 'stat',
         'field_name': 'stat',
         'prefix': ''},
        {'function_name': 'main_node',
         'field_name': 'main_node',
         'prefix': ''},
        {'function_name': 'neigh1',
         'field_name': 'neigh1',
         'prefix': ''},
        {'function_name': 'low',
         'field_name': 'lo',
         'prefix': ''},
        {'function_name': 'high',
         'field_name': 'hi',
         'prefix': ''},
        {'function_name': 'storage',
         'field_name': 'storage',
         'prefix': ''},
        {'function_name': 'data',
         'field_name': 'data',
         'prefix': ''},
        {'function_name': 'joined_neighbor',
         'field_name': 'neigh2',
         'prefix': ''},
        {'function_name': 'result',
         'field_name': 'result)',
         'prefix': 'aload(&'},
        {'function_name': 'left',
         'field_name': 'left)',
         'prefix': 'aload(&'},
        {'function_name': 'right',
         'field_name': 'right)',
         'prefix': 'aload(&'},
        {'function_name': 'key',
         'field_name': 'key',
         'prefix': ''},
        {'function_name': 'status',
         'field_name': 'status)',
         'prefix': 'aload(&'},
        {'function_name': 'root',
         'field_name': 'root)',
         'prefix': 'aload(&'},
    ],
    'to_assignments': [
        {'function_name': 'set_grand_parent_join_main',
         'field_name': 'gparent',
         'prefix': ''},
        {'function_name': 'set_other_branch_join_main',
         'field_name': 'otherb',
         'prefix': ''},
        {'function_name': 'set_neigh1_join_main',
         'field_name': 'neigh1',
         'prefix': ''}
    ],
    'remove_conversion_calls': [
        {'function_name': 'to_uintptr_t'},
        {'function_name': 'to_nodeptr'},
        {'function_name': 'route_status_to_uintptr_t'}
    ]
}

def find_parantecis_end(paran_start, text):
    opening = 1
    ending = 0
    current_pos = paran_start + 1
    while opening != ending:
        if text[current_pos] == '(':
            opening = opening + 1
        elif text[current_pos] == ')':
            ending = ending + 1
        current_pos = current_pos + 1
    return current_pos - 1

def find_parameter_end(param_start, text):
    current_pos = param_start
    while text[current_pos] != ',' and text[current_pos] != ')':
        if text[current_pos] == '(':
            current_pos = find_parantecis_end(current_pos, text) + 1
        else:
            current_pos = current_pos + 1
    return current_pos

# Returns a list of fuction calls to the function of the given name
# as well as a list of parameters
def find_calls(function_name, text):
    find_pattern = r"[^a-zA-Z1-9\_](" + function_name +"\s*\()"
    matches = re.finditer(find_pattern, text, re.DOTALL)
    output = []
    for match in matches:
        call_start = match.start(1)
        paran_start = match.end(1)
        call_end = find_parantecis_end(paran_start, text)
        parameters = []
        current_pos = paran_start
        while current_pos < call_end:
            param_end = find_parameter_end(current_pos, text)
            parameters.append(text[current_pos:param_end])
            current_pos = param_end + 1
        output.append((text[call_start:(call_end+1)],call_start, call_end, parameters))
    return output

def replace_calls(function_name, new_function_name, new_number_of_parameters, text):
    calls = find_calls(function_name, text)
    if len(calls) == 0:
        return text
    else:
        (call_text, call_start, call_end, parameters) = calls[0]
        new_call_text = new_function_name + "(" + ",".join(parameters[0:new_number_of_parameters]) + ")"
        transformed_text = text[:call_start] + new_call_text + text[call_end+1:]
        return transformed_text[:call_start+1] + replace_calls(function_name, new_function_name, new_number_of_parameters, transformed_text[call_start+1:])


def to_infix_field(function_name, field_name, prefix, text):
    calls = find_calls(function_name, text)
    if len(calls) == 0:
        return text
    else:
        (call_text, call_start, call_end, parameters) = calls[0]
        if len(parameters) != 1:
            raise Exception("Wrong number of paramters")
        new_call_text = prefix + parameters[0] + "->" + field_name
        transformed_text = text[:call_start] + new_call_text + text[call_end+1:]
        return transformed_text[:call_start+1] + to_infix_field(function_name, field_name, prefix, transformed_text[call_start+1:])


def to_assignment(function_name, field_name, prefix, text):
    calls = find_calls(function_name, text)
    if len(calls) == 0:
        return text
    else:
        (call_text, call_start, call_end, parameters) = calls[0]
        if len(parameters) != 2:
            raise Exception("Wrong number of paramters")
        new_call_text = prefix + parameters[0] + "->" + field_name + " =" + parameters[1]
        transformed_text = text[:call_start] + new_call_text + text[call_end+1:]
        return transformed_text[:call_start+1] + to_assignment(function_name, field_name, prefix, transformed_text[call_start+1:])

def remove_conversion_func(function_name, text):
    calls = find_calls(function_name, text)
    if len(calls) == 0:
        return text
    else:
        (call_text, call_start, call_end, parameters) = calls[0]
        if len(parameters) != 1:
            raise Exception("Wrong number of paramters")
        new_call_text = parameters[0]
        transformed_text = text[:call_start] + new_call_text + text[call_end+1:]
        return transformed_text[:call_start+1] + remove_conversion_func(function_name, transformed_text[call_start+1:])


def pick_out_parts(file_name):
    input_file = open(file_name, 'r')
    content = input_file.read()
    input_file.close()
    matches = re.finditer(configuration['pick_out_pattern'], content, re.DOTALL)
    output = []
    for match in matches:
        output.append((match.group(1), match.group(2)))
    return output

def transform_text(text):
    original_unannotated_text = text
    transformed_text = text
    # Do in-code replacements
    matches = re.finditer(configuration['replace_pattern'], text, re.DOTALL)
    for match in matches:
        to_replace = match.group(0)
        original = match.group(1) + match.group(2) + match.group(3)
        transformed = match.group(1) + match.group(4)
        original_unannotated_text = original_unannotated_text.replace(to_replace, original)
        transformed_text = transformed_text.replace(to_replace, transformed)
    # Do removes
    matches = re.finditer(configuration['remove_line_pattern'], transformed_text, re.DOTALL)  
    for match in matches:
        to_replace = match.group(1)
        original = match.group(2) + "\n"
        original_unannotated_text = original_unannotated_text.replace(to_replace, original)
        before_transe = transformed_text
        transformed_text = transformed_text.replace(to_replace, "", 1)
    # Do global replacements
    for replace_fun_call in configuration['replace_function_calls']:
        transformed_text = replace_calls(replace_fun_call['function_name'],
                                         replace_fun_call['new_function_name'],
                                         replace_fun_call['new_number_of_parameters'],
                                         transformed_text)
    for to_infix_field_desc in configuration['to_infix_fields']:
        transformed_text = to_infix_field(to_infix_field_desc['function_name'],
                                          to_infix_field_desc['field_name'],
                                          to_infix_field_desc['prefix'],
                                          transformed_text)
    for to_assignments_desc in configuration['to_assignments']:
        transformed_text = to_assignment(to_assignments_desc['function_name'],
                                         to_assignments_desc['field_name'],
                                         to_assignments_desc['prefix'],
                                         transformed_text)
    for remove_conversion_call in configuration['remove_conversion_calls']:
        transformed_text = remove_conversion_func(remove_conversion_call['function_name'],
                                                  transformed_text) 
    return (transformed_text, original_unannotated_text)

parts = pick_out_parts(configuration['input_file'])

for (name, text) in parts:
    transformed_text, original_unannotated_text = transform_text(text)
    part_file = open(name, 'w')
    part_file.write(transformed_text)
    part_file.close()
    if configuration['review_transformations'] and transformed_text != original_unannotated_text:
        org_unanot_file_name = name + ".org"
        org_unanot_file = open(org_unanot_file_name, 'w')
        org_unanot_file.write(original_unannotated_text)
        org_unanot_file.close()
        call([configuration['difftool'], org_unanot_file_name, name])

