/*
 *  Copyright 2018 Kjell Winblad (kjellwinblad@gmail.com, http://winsh.me)
 *
 *  This file is part of LFCA tree tests.
 *
 *  catrees is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  catrees is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with catrees.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <pthread.h>
#include <stdatomic.h>
#include <stdbool.h>
#include <stdlib.h>
#include "random.h"
#include <math.h>

pthread_mutex_t deterministic_scheduler_lock;

typedef struct {
    bool dead;
    bool running;
    bool is_global_locked;
    pthread_cond_t run_condition;
} deterministic_scheduler_thread_info;

#define MAX_NUMBER_OF_DETERMINISTICS_SCHEDULER_THREADS 128
deterministic_scheduler_thread_info deterministic_scheduler_thread_infos[MAX_NUMBER_OF_DETERMINISTICS_SCHEDULER_THREADS];
__thread int determistic_scheduler_thread_id;
int number_of_registered_deterministic_scheduler_threads = 0;
int number_of_running_threads = 0;
int number_of_alive_threads = 0;

pthread_t detsched_scheduler_thread;
pthread_cond_t schedule_condition;
unsigned int random_seed;

void  * scheduler_thread(void * not_used){
    (void)not_used;
    //sleep
    struct timespec sleep_time = {.tv_sec = 1, .tv_nsec = 100000000};
    nanosleep(&sleep_time, NULL);
    while(true){
        pthread_mutex_lock(&deterministic_scheduler_lock);    
        if(number_of_alive_threads == 0){
            pthread_mutex_unlock(&deterministic_scheduler_lock);
            break;
        }else if(number_of_running_threads == 0){
            // Wake up one thread at random
            int thread_id;
            do{
                thread_id = (number_of_registered_deterministic_scheduler_threads -1)*round(random_double(&random_seed));
            }while(deterministic_scheduler_thread_infos[thread_id].dead);
            deterministic_scheduler_thread_infos[thread_id].running = true;
            number_of_running_threads++;
            pthread_cond_signal(&deterministic_scheduler_thread_infos[thread_id].run_condition);
            pthread_mutex_unlock(&deterministic_scheduler_lock);
        }else{
            pthread_cond_wait(&schedule_condition, &deterministic_scheduler_lock);
            pthread_mutex_unlock(&deterministic_scheduler_lock);
        }
    }
    return NULL;
}


int detsched_pthread_create(pthread_t *thread, const pthread_attr_t *attr,
                            void *(*start_routine) (void *), void *arg){
    //register the thread
    pthread_mutex_lock(&deterministic_scheduler_lock);
    if(number_of_registered_deterministic_scheduler_threads == 0){
        random_seed = 43;
        pthread_create(&detsched_scheduler_thread, NULL, scheduler_thread, NULL);
    }
    if(number_of_registered_deterministic_scheduler_threads > MAX_NUMBER_OF_DETERMINISTICS_SCHEDULER_THREADS){
        exit(-1);
    }
    determistic_scheduler_thread_id = number_of_registered_deterministic_scheduler_threads;
    deterministic_scheduler_thread_infos[determistic_scheduler_thread_id].is_global_locked = false;
    deterministic_scheduler_thread_infos[determistic_scheduler_thread_id].dead = false;
    deterministic_scheduler_thread_infos[determistic_scheduler_thread_id].running = true;
    pthread_cond_init(&deterministic_scheduler_thread_infos[determistic_scheduler_thread_id].run_condition, NULL);
    number_of_registered_deterministic_scheduler_threads++;
    number_of_running_threads++;
    pthread_mutex_unlock(&deterministic_scheduler_lock);
    return pthread_create(thread, attr, start_routine, arg);
}

void detsched_before_shared_memory_access(){
    pthread_mutex_lock(&deterministic_scheduler_lock);
    number_of_running_threads--;
    pthread_cond_wait(&deterministic_scheduler_thread_infos[determistic_scheduler_thread_id].run_condition, &deterministic_scheduler_lock);
    pthread_mutex_unlock(&deterministic_scheduler_lock);
}

void detsched_before_lock_access(){
    pthread_cond_wait(&deterministic_scheduler_thread_infos[determistic_scheduler_thread_id].run_condition, &deterministic_scheduler_lock);
}

void detsched_before_lock_release(){

}


#define atomic_store(a, b) detsched_atomic_store(a, b)

#define atomic_load(a) detsched_atomic_load(a)

#define atomic_compare_exchange_strong(a, b, c) detsched_atomic_compare_exchange_strong(a, b, c)

#define pthread_mutex_lock(a) detsched_pthread_mutex_lock(a)
