#include <stdbool.h>

typedef struct treap_impl treap;

treap * treap_create();


/*
 * input:
 * m = input treap instance (will not be modified)
 * i = item to insert (must be positive)
 *
 * output:
 * treap = a treap instance containing the items of m + i
 * r_writeback = a treap instance containing the items of m + i
 * r_writeback = false if an item with value i altready existed and false otherwise
 */
treap * treap_insert(treap * m, int i, bool* r_writeback);

/*
 * input:
 * m = input treap instance (will not be modified)
 * i = item to insert (must be positive)
 *
 * output:
 * treap = a treap instance containing the items of m + i
 * r_writeback = a treap instance containing the items of m + i
 * r_writeback = false if an item with value i altready existed and false otherwise
 */
treap * treap_remove(treap * m, int i, bool* r_writeback);


/*
 * input:
 * m = input treap instance
 * i = item to lookup
 *
 * output:
 * true if the value i exist in i and false otherwise
 */
bool treap_lookup(treap * m, int i);

/*
 * input:
 * m = input treap instance
 *
 * output:
 * The key that will be used for splitting
 */
int treap_split_key(treap * m);


/*
 * input:
 * m = input treap instance
 *
 * output:
 * The left part of the tree when splitting
 */
treap * treap_split_left(treap * m);


/*
 * input:
 * m = input treap instance
 *
 * output:
 * The right part of the tree when splitting
 */
treap * treap_split_right(treap * m);

bool less_than_two_items(treap * m);

treap * treap_join(treap * t1, treap * t2);

/*
 * input:
 * m = input treap instance
 * traverser = a function that will be called once for all items in the range in icreasing order
 *                v = the current item in the range
 *                inaux = will be set to the value passed as the parameter aux
 * aux = some auxallary data (can be used to store items in range)
 *
 * output:
 * true if the value i exist in i and false otherwise
 */
void treap_range_query(treap * m, int start, int end, void (*traverser)(int v, void* inaux), void* aux);

void treap_print_range(treap * m, int start, int end);

void treap_check_equal(treap * m1, treap * m2, int start, int end);

void treap_print_dot_text(treap * t);

bool treap_is_empty(treap * m);

bool treap_max_key(treap * m);
