typedef struct lfcatree_impl lfcatree;
#include <stdbool.h>

lfcatree * lfcatree_create();

/*
 * input:
 * m = input treap instance (will not be modified)
 * i = item to insert (must be positive)
 *
 * output:
 * false if an item with value i already existed and false otherwise
 */
bool lfcatree_insert(lfcatree * m, int i);

/*
 * input:
 * m = input treap instance (will not be modified)
 * i = item to insert (must be positive)
 *
 * output:
 * false if an item with value i already existed and false otherwise
 */
bool lfcatree_remove(lfcatree * m, int i);


/*
 * input:
 * m = input treap instance
 * i = item to lookup
 *
 * output:
 * true if the value i exist in i and false otherwise
 */
bool lfcatree_lookup(lfcatree * m, int i);

/*
 * input:
 * m = input treap instance
 * traverser = a function that will be called once for all items in the range in icreasing order
 *                v = the current item in the range
 *                inaux = will be set to the value passed as the parameter aux
 * aux = some auxallary data (can be used to store items in range)
 *
 * output:
 * true if the value i exist in i and false otherwise
 */
void lfcatree_range_query(lfcatree * m, int start, int end, void (*traverser)(int v, void* inaux), void* aux);


void lfcatree_lookup_check(lfcatree * m);

void lfcatree_range_query_check_model(lfcatree * m, int start, int end);

void lfcatree_print_dot_text(lfcatree * t);
void low_contention_adaption_at(lfcatree * t, int key);

void high_contention_adaption_at(lfcatree * t, int key);
