/*
 *  Copyright 2018 Kjell Winblad (kjellwinblad@gmail.com, http://winsh.me)
 *
 *  This file is part of LFCA tree tests.
 *
 *  catrees is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  catrees is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with catrees.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <pthread.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include "immutable_treap.h"

typedef struct {
    pthread_mutex_t lock;
    treap * data;
} model;


model * model_create(){
    model * m = malloc(sizeof(model));
    pthread_mutex_init(&m->lock, NULL);
    m->data = NULL;
    return m;
}

/*
 * input:
 * m = input model instance (will not be modified)
 * i = item to insert (must be positive)
 *
 * output:
 * false if an item with value i already existed and false otherwise
 */
bool model_insert(model * m, int i){
    bool result;
    m->data = treap_insert(m->data, i, &result);
    return result;
}

/*
 * input:
 * m = input model instance (will not be modified)
 * i = item to insert (must be positive)
 *
 * output:
 * false if an item with value i already existed and false otherwise
 */
bool model_remove(model * m, int i){
    bool result;
    m->data = treap_remove(m->data, i, &result);
    return result;
}


/*
 * input:
 * m = input model instance
 * i = item to lookup
 *
 * output:
 * true if the value i exist in i and false otherwise
 */
bool model_lookup(model * m, int i){
    return treap_lookup(m->data, i);
}

/*
 * input:
 * m = input model instance
 * traverser = a function that will be called once for all items in the range in icreasing order
 *                v = the current item in the range
 *                inaux = will be set to the value passed as the parameter aux
 * aux = some auxallary data (can be used to store items in range)
 *
 * output:
 * true if the value i exist in i and false otherwise
 */
void model_range_query(model * m, int start, int end, void (*traverser)(int v, void* inaux), void* aux){
    treap_range_query(m->data, start, end, traverser, aux);
}

