#include "model.h"
#include "lock_free_catree.h"


/*
 * input:
 * X = input ordered set instance
 * i = item to insert
 *
 * output:
 * modifies X inplace
 * returns false if an item with value i already existed and true otherwise
 */
#define ordset_insert(X, i)                                             \
    _Generic((X),                                                       \
             model * : model_insert((model *)X, i),                     \
             lfcatree * : lfcatree_insert((lfcatree *)X, i))

/*
 * input:
 * X = input ordered set instance
 * i = item to remove
 *
 * output:
 * modifies X inplace
 * returns true if an item with value i existed and false otherwise
 */
#define ordset_remove(X, i)                                             \
    _Generic((X),                                                       \
             model * : model_remove((model *)X, i),                     \
             lfcatree * : lfcatree_remove((lfcatree *)X, i))

/*
 * input:
 * X = input ordered set instance
 * i = item to lookup
 *
 * output:
 * true if the value i existed in the set and false otherwise
 */
#define ordset_lookup(X, i)                                             \
    _Generic((X),                                                       \
             model * : model_lookup((model *)X, i),                     \
             lfcatree * : lfcatree_lookup((lfcatree *)X, i))


/*
 * input:
 * X = input ordered set instance
 * traverser = a function that will be called once for all items in the range in icreasing order
 *                v = the current item in the range
 *                inaux = will be set to the value passed as the parameter aux
 * aux = some auxallary data (can be used to store items in range)
 */
#define ordset_range_query(X, start, end, traverser, aux)               \
    _Generic((X),                                                       \
             model * : model_range_query((model *)X, start, end, traverser, aux), \
             lfcatree * : lfcatree_range_query((lfcatree *)X, start, end, traverser, aux))
