/*
 *  Copyright 2018 Kjell Winblad (kjellwinblad@gmail.com, http://winsh.me)
 *
 *  This file is part of LFCA tree tests.
 *
 *  catrees is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  catrees is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with catrees.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "lock_free_catree.h"
#include <pthread.h>
#include <stdatomic.h>
#include <stdio.h>
#include <assert.h>

#define DO_NOT_PRINT_TEST_NAME true

typedef  enum {
  test_name,//
  print_tree,
  ins,//
  multi_ins,
  rem,//
  get,//
  query,//
  split,//
  join,//
  PARALLEL_PART_START,//
  PARALLEL_PART_END,//
  END,//
  END_OF_SCENARIOS,//
} optype;


typedef struct {
  optype type;
  int param[2];
  char * name;
} op;

op scenarios[100][100] = {
  {
    {.type = test_name, .name = "Split at root"},
    {.type = multi_ins, .param = {2, 3}},
    {.type = PARALLEL_PART_START},
    {.type = split, .param = {2}},
    {.type = ins, .param = {0}},
    {.type = ins, .param = {4}},
    {.type = END},
  },
  /* { */
  /*   {.type = test_name, .name = "Join at root"}, */
  /*   {.type = multi_ins, .param = {2, 3}}, */
  /*   {.type = split, .param = {2}}, */
  /*   {.type = PARALLEL_PART_START}, */
  /*   {.type = join, .param = {0}}, */
  /*   {.type = ins, .param = {0}}, */
  /*   {.type = ins, .param = {4}}, */
  /*   {.type = END}, */
  /* }, */
    /* { */
    /*   {.type = test_name, .name = "Split at root range query"}, */
    /*   {.type = multi_ins, .param = {2, 3}}, */
    /*   {.type = PARALLEL_PART_START}, */
    /*   {.type = split, .param = {2}}, */
    /*   {.type = ins, .param = {0}}, */
    /*   {.type = ins, .param = {4}}, */
    /*   {.type = query, .param = {2,4}}, */
    /*   {.type = END}, */
    /* }, */
  /* { */
  /*   {.type = test_name, .name = "Join at root"}, */
  /*   {.type = multi_ins, .param = {2, 3}}, */
  /*   {.type = split, .param = {2}}, */
  /*   {.type = PARALLEL_PART_START}, */
  /*   {.type = join, .param = {2}}, */
  /*   {.type = ins, .param = {0}}, */
  /*   {.type = ins, .param = {4}}, */
  /*   //{.type = query, .param = {0,1}}, */
  /*   {.type = END}, */
  /* }, */
  {
    {.type = END_OF_SCENARIOS}
  }
};

typedef struct {
  int param[2];
  lfcatree * tree;
} op_info;


void  * insert_thread(void * param){
  op_info * o = param;
  //printf("INSERT START\n");
  lfcatree_insert(o->tree, o->param[0]);
  //printf("INSERT END\n");
  return NULL;
}

void  * remove_thread(void * param){
  op_info * o = param;
  lfcatree_remove(o->tree, o->param[0]);
  return NULL;
}

void  * lookup_thread(void * param){
  op_info * o = param;
  lfcatree_lookup(o->tree, o->param[0]);
  return NULL;
}

void  * range_query_thread(void * param){
  op_info * o = param;
  //printf("RQ START\n");
  lfcatree_range_query_check_model(o->tree, o->param[0], o->param[1]);
  //printf("RQ END\n");
  return NULL;
}

void  * split_thread(void * param){
  op_info * o = param;
  //printf("SPLIT START\n");
  high_contention_adaption_at(o->tree, o->param[0]);
  //printf("SPLIT END\n");
  return NULL;
}

void  * join_thread(void * param){
  op_info * o = param;
  low_contention_adaption_at(o->tree, o->param[0]);
  return NULL;
}


int main (int argc, char *argv[]) {
  //printf("START\n");
  pthread_t threads[128];
    int number_of_threads = 0;
    int scenario_id = 0;
    while(true){
      op* operations = scenarios[scenario_id];
      int operation_id = 0;
      bool parallel_part = false;
      op o = operations[operation_id];
      lfcatree * tree = lfcatree_create();
      while(o.type != END){
        //printf("LOOP\n"); 
        if(o.type == END_OF_SCENARIOS){
          assert(number_of_threads == 0);
          return 0;
        } else if(o.type == PARALLEL_PART_START){
          parallel_part = true;
          assert(number_of_threads == 0);
        } else if(o.type == PARALLEL_PART_END){
          int threads_to_remove = number_of_threads;
          for(int i = 0; i < threads_to_remove; i++){
            void* ignore;
            pthread_join(threads[i], &ignore);
            number_of_threads--;
          }
        } else if(o.type == test_name){
#ifndef DO_NOT_PRINT_TEST_NAME
          printf("%s\n", o.name);
#endif
        }else if(!parallel_part){
          if(o.type == ins){
            lfcatree_insert(tree, o.param[0]);
          }else if(o.type == multi_ins){
            for(int i = o.param[0]; i <= o.param[1]; i++){
              lfcatree_insert(tree, i);
            }
          }else if(o.type == rem){
            lfcatree_remove(tree, o.param[0]);
          }else if(o.type == get){
            lfcatree_lookup(tree, o.param[0]);
          }else if(o.type == query){
            lfcatree_range_query_check_model(tree, o.param[0], o.param[1]);
          }else if(o.type == split){
            high_contention_adaption_at(tree, o.param[0]);
          }else if(o.type == join){
            low_contention_adaption_at(tree, o.param[0]);
          }else if(o.type == print_tree){
            lfcatree_print_dot_text(tree);
          }
        }else{
          op_info info = {.param[0] = o.param[0], .param[1] = o.param[1], .tree = tree};
          if(o.type == ins){
            pthread_create(&threads[number_of_threads], NULL, insert_thread, &info);
            number_of_threads++; 
          }else if(o.type == multi_ins){
            for(int i = o.param[0]; i <= o.param[1]; i++){
              info.param[0] = i;
              pthread_create(&threads[number_of_threads], NULL, insert_thread, &info);
              number_of_threads++; 
            }
          }else if(o.type == rem){
            pthread_create(&threads[number_of_threads], NULL, remove_thread, &info);
            number_of_threads++; 
          }else if(o.type == get){
            pthread_create(&threads[number_of_threads], NULL, lookup_thread, &info);
            number_of_threads++; 
          }else if(o.type == query){
            //printf("RANGE Q THREAD\n");
            pthread_create(&threads[number_of_threads], NULL, range_query_thread, &info);
            number_of_threads++; 
          }else if(o.type == split){
            pthread_create(&threads[number_of_threads], NULL, split_thread, &info);
            number_of_threads++; 
          }else if(o.type == join){
            pthread_create(&threads[number_of_threads], NULL, join_thread, &info);
            number_of_threads++; 
          }
        }
        operation_id++;
        o = operations[operation_id];
      }
      if(number_of_threads > 0){
        int threads_to_remove = number_of_threads;
        for(int i = 0; i < threads_to_remove; i++){
          void* ignore;
          pthread_join(threads[i], &ignore);
          number_of_threads--;
        }
      }
      scenario_id = scenario_id + 1;
    }
    //printf("E\n");
    return 0;
}
