/*
 *  Copyright 2018 Kjell Winblad (kjellwinblad@gmail.com, http://winsh.me)
 *
 *  This file is part of LFCA tree tests.
 *
 *  catrees is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  catrees is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with catrees.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "test_framework.h"
#include "immutable_treap.h"
#include <stdlib.h>


int test_create(){
    treap * treap = treap_create();
    return 1;
}

int test_insert(int number_of_inserts){
    treap * treap = treap_create();
    for(int i = 0; i < number_of_inserts; i++){
        bool res;
        treap = treap_insert(treap, i, &res);
        ASSERT(res, "insert should succeed");
    }
    for(int i = 0; i < number_of_inserts; i++){
        bool res;
        treap = treap_insert(treap, i, &res);
        ASSERT(!res, "insert should fail");
    }
    return 1;
}

int test_remove(int number_of_removes){
    treap * treap = treap_create();
    for(int i = 0; i < number_of_removes; i = i + 2){
        bool res;
        treap = treap_insert(treap, i, &res);
        ASSERT(res, "insert should succeed");
    }
    for(int i = 0; i < number_of_removes; i++){
        bool res;
        treap = treap_remove(treap, i, &res);
        if(i%2 == 0){
            ASSERT(res, "remove should succseed");
        }else{
            ASSERT(!res, "remove should fail");
        }
    }
    ASSERT(treap == NULL, "treap should be NULL now")
    return 1;
}

int test_lookup(int number_of_removes){
    treap * treap = treap_create();
    for(int i = 0; i < number_of_removes; i = i + 2){
        bool res;
        treap = treap_insert(treap, i, &res);
        ASSERT(res, "insert should succeed");
    }
    for(int i = 0; i < number_of_removes; i++){
        bool res;
        res = treap_lookup(treap, i);
        if(i%2 == 0){
            ASSERT(res, "lookup should succseed");
        }else{
            ASSERT(!res, "lookup should fail");
        }
    }
    ASSERT(treap != NULL, "treap should not be NULL now")
    return 1;
}

    
typedef struct {
    int size;
    int capacity;
    int * array;
} range_query_result_buffer;

range_query_result_buffer * result_empty(){
    range_query_result_buffer * res = malloc(sizeof(range_query_result_buffer));
    res->size=0;
    res->capacity=16;
    res->array = malloc(16*sizeof(int));
    return res;
}


int result_size(range_query_result_buffer * res){
    if(res == NULL){
        return 0;
    }else{
        return res->size;
    }
}

range_query_result_buffer * result_add(range_query_result_buffer * res, int new_item){
    if(res == NULL){
        res = malloc(sizeof(range_query_result_buffer));
        res->size = 0;
        res->capacity = 1;
        res->array = malloc(sizeof(int));
        return result_add(res, new_item);
    }else{
        if(res->size == res->capacity){
            int * new_array = malloc(sizeof(int) * (res->capacity *2));
            for(int i = 0; i < res->size; i++){
                new_array[i] = res->array[i];
            }
            res->array = new_array;
            res->capacity = res->capacity*2;
        }
        res->array[res->size] = new_item;
        res->size = res->size + 1;
        return res;
    }
}

void range_query_test_traverser(int v, void* inaux){
    range_query_result_buffer * res = (range_query_result_buffer *)inaux;
    result_add(res, v);
}


int max(int m1, int m2){
    return m1 > m2 ? m1 : m2;
}

int min(int m1, int m2){
    return m1 < m2 ? m1 : m2;
}

int test_range(treap * treap, int start, int end){
    range_query_result_buffer * res = result_empty();
    int range_size = 0;
    treap_range_query(treap, start, end,  range_query_test_traverser, res);
    for(int i = max(0,start); i <= min(998, end); i++){
        if(i % 2 == 0){
            ASSERT(res->array[range_size] == i, "Should be in range");
            range_size++;
        }
    }
    ASSERT(res->size == range_size, "Should be the range size");
    return 1;
}

int test_range_query(){
    treap * treap = treap_create();
    for(int i = 0; i < 1000; i = i + 2){
        bool res;
        treap = treap_insert(treap, i, &res);
        ASSERT(res, "insert should succeed");
    }
    range_query_result_buffer * res = result_empty();
    test_range(treap, -10, 10);
    test_range(treap, 1, 10);
    test_range(treap, 2, 11);
    test_range(treap, -10, 2000);
    test_range(treap, -10, 998);
    test_range(treap, 6, 6);
    test_range(treap, 7, 7);
    test_range(treap, 7, 8);
    test_range(treap, 996, 1000);
    return 1;
}

int main (int argc, char *argv[]) {
    int number = 0;
    T(test_create(), "test_create()");
    T(test_insert(1), "test_insert(1)");
    T(test_insert(10), "test_insert(10)");
    T(test_insert(10001), "test_insert(10001)");
    T(test_remove(1), "test_remove(1)");
    T(test_remove(10), "test_remove(10)");
    T(test_remove(100), "test_remove(100)");
    T(test_remove(10001), "test_remove(10001)");
    T(test_lookup(1), "test_loookup(1)");
    T(test_lookup(10), "test_loookup(10)");
    T(test_lookup(100), "test_loookup(100)");
    T(test_lookup(1000), "test_loookup(1000)");
    T(test_lookup(10001), "test_loookup(10001)");
    T(test_range_query(), "test_range_query()");
    return 0;
} 
