/*
 *  Copyright 2018 Kjell Winblad (kjellwinblad@gmail.com, http://winsh.me)
 *
 *  This file is part of LFCA tree tests.
 *
 *  catrees is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  catrees is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with catrees.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "test_framework.h"
#include "ordered_set.h"
#include "tools.h"
#include <stdlib.h>


#ifdef TEST_MODEL
#define CONSTRUCTOR() model_create()
#define TYPE model
#else
#define CONSTRUCTOR() lfcatree_create()
#define TYPE lfcatree
#endif


int test_create(){
    TYPE * treap = CONSTRUCTOR();
    return 1;
}

int test_insert(int number_of_inserts){
    TYPE * treap = CONSTRUCTOR();
    for(int i = 0; i < number_of_inserts; i++){
        bool res = ordset_insert(treap, i);
        ASSERT(res, "insert should succeed");
    }
    for(int i = 0; i < number_of_inserts; i++){
        bool res = ordset_insert(treap, i);
        ASSERT(!res, "insert should fail");
    }
    return 1;
}

int test_remove(int number_of_removes){
    TYPE * treap = CONSTRUCTOR();
    for(int i = 0; i < number_of_removes; i = i + 2){
        bool res = ordset_insert(treap, i);
        ASSERT(res, "insert should succeed");
    }
    for(int i = 0; i < number_of_removes; i++){
        bool res = ordset_remove(treap, i);
        if(i%2 == 0){
            ASSERT(res, "remove should succseed");
        }else{
            ASSERT(!res, "remove should fail");
        }
    }
    return 1;
}

int test_lookup(int number_of_removes){
    TYPE * treap = CONSTRUCTOR();
    for(int i = 0; i < number_of_removes; i = i + 2){
        bool res = ordset_insert(treap, i);
        ASSERT(res, "insert should succeed");
    }
    for(int i = 0; i < number_of_removes; i++){
        bool res = ordset_lookup(treap, i);
        if(i%2 == 0){
            ASSERT(res, "lookup should succseed");
        }else{
            ASSERT(!res, "lookup should fail");
        }
    }
    return 1;
}

void range_query_test_traverser(int v, void* inaux){
    range_query_result_buffer * res = (range_query_result_buffer *)inaux;
    result_add(res, v);
}

int test_range(TYPE * treap, int start, int end){
    range_query_result_buffer * res = result_empty();
    int range_size = 0;
    ordset_range_query(treap, start, end, range_query_test_traverser, res);
    for(int i = max(0,start); i <= min(998, end); i++){
        if(i % 2 == 0){
            ASSERT(res->array[range_size] == i, "Should be in range");
            range_size++;
        }
    }
    ASSERT(res->size == range_size, "Should be the range size");
    return 1;
}

int test_range_query(){
    TYPE * treap = CONSTRUCTOR();
    for(int i = 0; i < 1000; i = i + 2){
        bool res = ordset_insert(treap, i);
        ASSERT(res, "insert should succeed");
    }
    range_query_result_buffer * res = result_empty();
    if(!test_range(treap, -10, 10)) return 0;
    if(!test_range(treap, 1, 10)) return 0;
    if(!test_range(treap, 2, 11)) return 0;
    if(!test_range(treap, -10, 2000)) return 0;
    if(!test_range(treap, -10, 998)) return 0;
    if(!test_range(treap, 6, 6)) return 0;
    if(!test_range(treap, 7, 7)) return 0;
    if(!test_range(treap, 7, 8)) return 0;
    if(!test_range(treap, 996, 1000)) return 0;
    return 1;
    
}

int main (int argc, char *argv[]) {
    T(test_create(), "test_create()");
    T(test_insert(1), "test_insert(1)");
    T(test_insert(10), "test_insert(10)");
    T(test_insert(10001), "test_insert(10001)");
    T(test_remove(1), "test_remove(1)");
    T(test_remove(10), "test_remove(10)");
    T(test_remove(100), "test_remove(100)");
    T(test_remove(10001), "test_remove(10001)");
    T(test_lookup(1), "test_loookup(1)");
    T(test_lookup(10), "test_loookup(10)");
    T(test_lookup(100), "test_loookup(100)");
    T(test_lookup(1000), "test_loookup(1000)");
    T(test_lookup(10001), "test_loookup(10001)");
    T(test_range_query(), "test_range_query()");
    return 0;
} 
